/***************************************************************
* Program : Calculation of PI by integration of 1/(1+x*x)
*           using MPI
* Project #2 -- CPS 615
*    
* Desc    : Program uses an approximation of the above integral
*           to calculate the value of PI. It then broadcasts this
*           value to other process, which display the value
*           independently. 
*
* Author  : William Gropp, ANL.
* 
* Modified by : Sachin R. Shanbhag.
*****************************************************************/


#include <stdio.h>
#include <mpi.h>
#include <math.h>

int main(int argc, char *argv[])
{
  int n, myid, numprocs, i,j, locarrcnt;
  double PI25DT = 3.14159265358979323842643;/*Const. value of pi*/
  
  double pi, h, x, sum;

  float glob_sum[50], locarr[50];        /* arrays to store sums*/

  /**************************************************************
   Initialise MPI and get no. of processes and my rank.
  ***************************************************************/

  MPI_Init(&argc, &argv);
  
  MPI_Comm_size(MPI_COMM_WORLD, &numprocs);
  
  MPI_Comm_rank(MPI_COMM_WORLD, &myid);
  locarrcnt = 0;

  n = 50; /* No. of summations */

  h = 1.0 /(double) n;  
  sum = 0.0;

  /**************************************************************
    Each process does a part of the total no. of calculations
  ***************************************************************/
  
  for (i=myid+1; i<=n; i+=numprocs){
     x = h *((double)i - 0.5);
     locarr[locarrcnt++] = (4.0/(1.0 + x*x))*h;
  }

  /**************************************************************
    The root process gathers all the calculated values.
    The Gather operation is used here. Notice the rank 0 specified
    as the second last paramter to indicate the root gathers all 
    the values even though the function is in the process space of
    all the processes.
  ***************************************************************/

  MPI_Gather(locarr, locarrcnt, MPI_FLOAT, glob_sum, locarrcnt, MPI_FLOAT, 0, MPI_COMM_WORLD);
  
  /***************************************************************
    The Root process then sums up all the values and displays it
    along wiht the error
  ***************************************************************/

  if (myid == 0){
     pi = 0.0;
     for(j=1; j<=n; j++){
	pi = pi+glob_sum[j-1];
     }
     printf("PI calculated on root (0) process: %.16f\n", pi);
     printf("Error in calculation :%16f\n     ", fabs(pi - PI25DT));
     printf("\rNow broadcasting value of PI to other processes ... \n\n");
  }

  /***************************************************************
    The root then broadcasts this value to all the processes which
    then display it in turn. Note that this function too is in the
    global process space. The root process rank is mentioned as the
    second-last parameter.
  ****************************************************************/

  MPI_Bcast(&pi, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
  printf("Message from %d : value of PI received is %.16f\n", myid, pi);
	
  MPI_Finalize;
}
